from __future__ import annotations

from datetime import datetime
import json
import os
from pathlib import Path
from typing import Any, Iterable, Optional

import click
import requests
import toml
from tqdm import tqdm
from xdg import BaseDirectory

from swcc.api import SwccSession


def raise_for_status(response: requests.Response):
    try:
        response.raise_for_status()
    except requests.HTTPError:
        data = None
        try:
            data = json.dumps(response.json(), indent=2)
        except Exception:
            data = response.text

        if data:
            click.echo(click.style(f'Received:\n{data}\n', fg='yellow'), err=True)
        raise


def download_file(r: requests.Response, dest: Path, name: str, mtime: Optional[datetime] = None):
    filename = dest / name
    with tqdm.wrapattr(
        open(filename, 'wb'),
        'write',
        miniters=1,
        desc=str(filename),
        total=int(r.headers.get('content-length', 0)),
    ) as f:
        for chunk in r.iter_content(1024 * 1024 * 16):
            f.write(chunk)

    if mtime:
        os.utime(filename, (datetime.now().timestamp(), mtime.timestamp()))


def files_to_upload(session: SwccSession, existing: Iterable, path: Path) -> Iterable[Path]:
    existing_filenames = set([e.name for e in existing])
    for filename in os.listdir(path):
        if filename not in existing_filenames:
            yield Path(path / filename)


def upload_path(session: SwccSession, path: Path, field: str, name=None) -> str:
    if name is None:
        name = path.name
    with path.open('rb') as f:
        return session.s3ff.upload_file(f, name, field)['field_value']


def update_config_value(filename: str, key: str, value: Any) -> None:
    from swcc import SWCC_CONFIG_PATH

    BaseDirectory.save_config_path(SWCC_CONFIG_PATH)

    config_dir = BaseDirectory.load_first_config(SWCC_CONFIG_PATH)
    config_file = os.path.join(config_dir, filename)

    if os.path.exists(config_file):
        with open(config_file, 'r') as infile:
            config = toml.load(infile)
            config['default'][key] = value
    else:
        config = {'default': {key: value}}

    with open(config_file, 'w') as outfile:
        toml.dump(config, outfile)


def get_config_value(filename: str, key: str) -> Optional[Any]:
    from swcc import SWCC_CONFIG_FILE, SWCC_CONFIG_PATH

    BaseDirectory.save_config_path(SWCC_CONFIG_PATH)
    config_dir: Optional[str] = BaseDirectory.load_first_config(SWCC_CONFIG_PATH)

    if config_dir:
        config_file = os.path.join(config_dir, SWCC_CONFIG_FILE)

        if os.path.exists(config_file):
            with open(config_file, 'r') as infile:
                config = toml.load(infile)['default']
                return config.get(key)

    return None



from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory

from factory import Factory, Faker, Sequence, SubFactory
from faker.providers import BaseProvider, file as file_provider, python as python_provider

from swcc import models


class FileUploadProvider(BaseProvider):
    _directory: Path

    def file(self, category=None, extension=None, null_chance=0):
        if self.generator.boolean(null_chance):
            return
        file_name = self.generator.file_name(extension=extension, category=category)
        file_path = self._directory / file_name
        with file_path.open('wb') as f:
            f.write(b' ')

        return file_path


Faker.add_provider(python_provider)
Faker.add_provider(file_provider)
Faker.add_provider(FileUploadProvider)


@contextmanager
def file_context():
    with TemporaryDirectory() as directory:
        p = Path(directory)
        FileUploadProvider._directory = p
        yield p

    del FileUploadProvider._directory


Faker.add_provider(FileUploadProvider)


class DatasetFactory(Factory):
    class Meta:
        model = models.Dataset

    name = Sequence(lambda n: f'dataset_{n}')
    license = Faker('sentence')
    description = Faker('sentence')
    acknowledgement = Faker('sentence')
    keywords = Faker('word')
    contributors = Faker('name')


class SubjectFactory(Factory):
    class Meta:
        model = models.Subject

    name = Sequence(lambda n: f'subject_{n}')
    dataset = SubFactory(DatasetFactory)


class SegmentationFactory(Factory):
    class Meta:
        model = models.Segmentation

    file = Faker('file', extension='nrrd')
    anatomy_type = Faker('word')
    subject = SubFactory(SubjectFactory)


class MeshFactory(Factory):
    class Meta:
        model = models.Mesh

    file = Faker('file', extension='ply')
    anatomy_type = Faker('word')
    subject = SubFactory(SubjectFactory)


class ImageFactory(Factory):
    class Meta:
        model = models.Image

    file = Faker('file', extension='nrrd')
    modality = Faker('word')
    subject = SubFactory(SubjectFactory)


class ProjectFactory(Factory):
    class Meta:
        model = models.Project

    file = Faker('file', extension='csv')
    keywords = Faker('word')
    description = Faker('sentence')
    dataset = SubFactory(DatasetFactory)


class GroomedSegmentationFactory(Factory):
    class Meta:
        model = models.GroomedSegmentation

    file = Faker('file', extension='nrrd')
    pre_cropping = Faker('file', extension='txt', null_chance=50)
    pre_alignment = Faker('file', extension='txt', null_chance=50)

    segmentation = SubFactory(SegmentationFactory)
    project = SubFactory(ProjectFactory)


class GroomedMeshFactory(Factory):
    class Meta:
        model = models.GroomedMesh

    file = Faker('file', extension='ply')
    pre_cropping = Faker('file', extension='txt', null_chance=50)
    pre_alignment = Faker('file', extension='txt', null_chance=50)

    mesh = SubFactory(MeshFactory)
    project = SubFactory(ProjectFactory)


class OptimizedShapeModelFactory(Factory):
    class Meta:
        model = models.OptimizedShapeModel

    project = SubFactory(ProjectFactory)
    parameters = Faker('pydict', value_types=[str, int, float])


class OptimizedParticlesFactory(Factory):
    class Meta:
        model = models.OptimizedParticles

    world = Faker('file', extension='txt')
    local = Faker('file', extension='txt')
    transform = Faker('file', extension='txt')
    shape_model = SubFactory(OptimizedShapeModelFactory)
    groomed_segmentation = SubFactory(GroomedSegmentationFactory)
    groomed_mesh = SubFactory(GroomedMeshFactory)


class OptimizedPCAModelFactory(Factory):
    class Meta:
        model = models.OptimizedPCAModel

    mean_particles = Faker('file', extension='txt')
    pca_modes = Faker('file', extension='txt')
    eigen_spectrum = Faker('file', extension='txt')
    shape_model = SubFactory(OptimizedShapeModelFactory)



from __future__ import annotations

import itertools
import logging
from pathlib import Path, PurePath
import re
from tempfile import TemporaryDirectory

try:
    from typing import (
        Any,
        Dict,
        Generic,
        Iterator,
        Literal,
        Optional,
        Tuple,
        Type,
        TypeVar,
        Union,
        get_args,
    )
except ImportError:
    from typing import (
        Any,
        Dict,
        Generic,
        Iterator,
        Optional,
        Tuple,
        Type,
        TypeVar,
        Union,
    )
    from typing_extensions import (  # type: ignore
        Literal,
        get_args,
    )

from urllib.parse import unquote

from openpyxl import load_workbook
from pydantic import (
    AnyHttpUrl,
    BaseModel,
    FilePath,
    StrictStr,
    ValidationError,
    parse_obj_as,
    validator,
)
from pydantic.fields import ModelField
import requests

from .api import current_session
from .utils import raise_for_status

logger = logging.getLogger(__name__)

FieldId = TypeVar('FieldId', bound=str)


def shape_file_type(path: Path) -> Type[Segmentation] | Type[Mesh]:
    """
    Determine the type of the shape file.

    As described in https://github.com/SCIInstitute/ShapeWorks/blob/3344bbfd42cb83eea50d01c06218e3034b5c67aa/Libs/Mesh/Mesh.h#L212,
    Meshes are of file type "vtk", "vtp", "ply", "stl", "obj", while Segmentations are any other
    file type.
    """  # noqa: E501
    file_type = path.suffix
    if file_type in ['.vtk', '.vtp', '.ply', '.stl', '.obj']:
        return Mesh
    else:
        return Segmentation


class NonEmptyString(StrictStr):
    min_length = 1


class FileType(Generic[FieldId]):
    def __init__(
        self,
        path: Optional[Union[Path, str]] = None,
        url: Optional[str] = None,
        field_id: Optional[str] = None,
    ):
        self.field_id: Optional[str] = field_id
        self.path: Optional[Path] = None
        self.url: Optional[AnyHttpUrl] = None
        self.field_value: Optional[str] = None

        if path:
            self.path = parse_obj_as(FilePath, path)

        if url:
            self.url = parse_obj_as(AnyHttpUrl, url)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v, field: ModelField, **kwargs):
        # this is probably a bad idea...
        if not field.sub_fields:
            raise SyntaxError('A field id must be provided when using FileType')

        field_id = get_args(field.sub_fields[0].type_)[0]
        if isinstance(v, FileType):
            v.field_id = field_id
            return v

        path = None
        url = None
        try:
            path = parse_obj_as(FilePath, v)
        except ValidationError:
            pass

        try:
            url = parse_obj_as(AnyHttpUrl, v)
        except ValidationError:
            pass

        if path is None and url is None:
            raise ValueError(f'Could not parse {v} as a local path or a remote url')

        return FileType(path=path, url=url, field_id=field_id)

    def upload(self):
        session = current_session()

        if self.field_value:
            return self.field_value

        if self.url is not None or self.path is None:
            # trying to upload a file when the model was sourced from the server
            raise Exception('Cannot upload a remote file reference')

        if self.field_id is None:
            # I assume validate will always get called, but pydantic *might* be doing something
            # unusual in some cases.
            raise Exception('Unknown field id, this is likely a bug in the FileType class')

        with self.path.open('rb') as f:
            logger.info('Uploading file %s', self.path.name)
            self.field_value = session.s3ff.upload_file(f, str(self.path.name), self.field_id)
            logger.debug('Uploaded file %s', self.path.name)

        return self.field_value

    def download(self, path: Union[Path, str]) -> Path:
        if self.url is None:
            raise Exception('Cannot download a local file')

        if not self.url.path:
            raise Exception('Server returned an unexpected url')

        path = Path(path)
        if path.is_file():
            raise ValueError('Expected a directory name')

        if not path.is_dir():
            path.mkdir(parents=True, exist_ok=True)

        path = path / self.url.path.split('/')[-1]
        r = requests.get(self.url, stream=True)
        raise_for_status(r)

        logger.info('Downloading %s', path)
        with path.open('wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
        return path

    @property
    def name(self) -> str:
        if self.path:
            return self.path.name
        elif self.url:
            if self.url.path is None:
                raise Exception('Invalid file url')
            return unquote(self.url.path.split('/')[-1])
        raise Exception('Invalid file object')

    def __str__(self):
        return self.name


ModelType = TypeVar('ModelType', bound='ApiModel')


class ApiModel(BaseModel):
    _endpoint: str

    id: Optional[int]

    @validator('*', pre=True)
    def fetch_entity(cls, v, field: ModelField):
        type_ = cls.__fields__[field.name].type_
        if type_ is not Any and issubclass(type_, ApiModel) and isinstance(v, int):
            return type_.from_id(v)
        return v

    @classmethod
    def from_id(cls: Type[ModelType], id: int, **kwargs) -> ModelType:
        session = current_session()
        cache = session.cache[cls]

        if id not in cache:
            r: requests.Response = session.get(f'{cls._endpoint}/{id}/')
            raise_for_status(r)
            json = r.json()
            for key, value in cls.__fields__.items():
                if key in kwargs:
                    json[key] = kwargs[key]
                elif value.type_ is not Any and issubclass(value.type_, ApiModel):
                    json[key] = value.type_.from_id(json[key])
            cache[id] = cls(**json)
        return cache[id]

    @classmethod
    def list(cls: Type[ModelType], **kwargs) -> Iterator[ModelType]:
        session = current_session()

        filter: Dict[str, Any] = {}
        replace: Dict[str, ApiModel] = {}
        for key, value in kwargs.items():
            if isinstance(value, ApiModel):
                filter[key] = value.id
                replace[key] = value
            else:
                filter[key] = value

        r: requests.Response = session.get(f'{cls._endpoint}/', params=filter)

        while True:
            raise_for_status(r)
            data = r.json()
            for result in data['results']:
                result.update(replace)
                yield cls(**result)
            if not data.get('next'):
                return

            r = session.get(data['next'])

    def delete(self) -> None:
        session = current_session()

        self.assert_remote()
        r: requests.Response = session.delete(f'{self._endpoint}/{self.id}/')
        raise_for_status(r)
        self.id = None

    def create(self: ModelType) -> ModelType:
        session = current_session()

        self.assert_local()
        json = self.dict()
        for key, value in self:
            if isinstance(value, FileType):
                json[key] = value.upload()
            if isinstance(value, ApiModel):
                if value.id is None:
                    value.create()
                json[key] = value.id

        r: requests.Response = session.post(f'{self._endpoint}/', json=json)
        raise_for_status(r)
        self.id = r.json()['id']
        return self

    def download_files(self, path: Union[Path, str]) -> Iterator[Path]:
        for _, value in self:
            if isinstance(value, FileType):
                yield value.download(path)

    def assert_remote(self):
        if self.id is None:
            raise Exception('This entity has not yet been created on the server.')

    def assert_local(self):
        if self.id is not None:
            raise Exception('This entity already exists on the server (id: %r).' % self.id)


class Dataset(ApiModel):
    _endpoint = 'datasets'

    name: NonEmptyString
    file: Optional[FileType[Literal['core.Dataset.file']]] = None
    license: NonEmptyString
    description: NonEmptyString
    acknowledgement: NonEmptyString
    keywords: str = ''
    contributors: str = ''
    publications: str = ''

    def __repr_args__(self: BaseModel):
        return [
            (key, value)
            for key, value in self.__dict__.items()
            if key
            not in {
                'license',
                'description',
                'acknowledgement',
                'keywords',
                'contributors',
                'publications',
            }
        ]

    @property
    def subjects(self) -> Iterator[Subject]:
        self.assert_remote()
        return Subject.list(dataset=self)

    def add_subject(self, name: str) -> Subject:
        return Subject(name=name, dataset=self).create()

    @property
    def projects(self) -> Iterator[Project]:
        self.assert_remote()
        return Project.list(dataset=self)

    @property
    def segmentations(self) -> Iterator[Segmentation]:
        for subject in self.subjects:
            for segmentation in subject.segmentations:
                yield segmentation

    @property
    def meshes(self) -> Iterator[Mesh]:
        for subject in self.subjects:
            for mesh in subject.meshes:
                yield mesh

    @property
    def images(self) -> Iterator[Image]:
        for subject in self.subjects:
            for mesh in subject.images:
                yield mesh

    def force_create(self, backup=False):
        """
        Forcibly create the Dataset, even if it already exists.

        If backup=False (the default), then the existing Dataset is deleted.
        If backup=True, then the new Dataset will append an appropriate `-v*` version string to
        its name before creation.
        """
        old_dataset = Dataset.from_name(self.name)
        while old_dataset is not None:
            if not backup:
                # Delete the old dataset to resolve the collision
                old_dataset.delete()
            else:
                # Use a version suffix to resolve the collision
                version_regex = re.compile('(.*)-v([0-9]+)')
                match = version_regex.match(self.name)
                if match:
                    # The old name had a suffix, so increment it
                    name, old_version = match.groups()
                    logger.info('Trying a new name: %s, %s', name, old_version)
                    new_version = int(old_version) + 1
                    self.name = f'{name}-v{new_version}'  # type: ignore
                else:
                    # The old name had no suffix, so append "-v1"
                    self.name = f'{self.name}-v1'  # type: ignore
            # We have a new name now, but that new name might also conflict.
            # Keep looping until there is no conflict.
            old_dataset = Dataset.from_name(self.name)
        return self.create()

    def add_project(self, file: Path, keywords: str = '', description: str = '') -> Project:
        project = Project(
            file=file,
            keywords=keywords,
            description=description,
            dataset=self,
        ).create()
        return project._load_project_spreadsheet()

    @classmethod
    def from_name(cls, name: str) -> Optional[Dataset]:
        results = cls.list(name=name)
        try:
            return next(results)
        except StopIteration:
            return None

    def create(self) -> Dataset:
        result = super().create()
        if self.file:
            self._load_data_spreadsheet()
        # Load the new dataset so we get an appropriate file field
        assert result.id
        return Dataset.from_id(result.id)

    def _load_data_spreadsheet(self):  # noqa: C901
        if not self.file or not self.file.path:
            return
        file = self.file.path

        xls = load_workbook(str(file), read_only=True)
        if 'data' not in xls:
            raise Exception('`data` sheet not found')

        data = xls['data'].values

        headers = next(data)
        for header in headers:
            if not (header.startswith('shape_') or header.startswith('image_')):
                raise Exception(
                    f'Unknown spreadsheet format in {file} - expected "shape_" or "image_" prefix, found {header}'  # noqa: E501
                )
        # split each header into ('shape', anatomy_type) and ('image', modality) tuples
        column_info = [header.split('_', 1) for header in headers]
        # at least one shape column must be present
        if all(info[0] != 'shape' for info in column_info):
            raise Exception('No "shape_" column specified')

        root = file.parent
        subjects: Dict[str, Subject] = {}

        found = False
        for row in data:
            subject = None
            for i, cell in enumerate(row):
                if not cell:
                    continue
                file_type, domain = column_info[i]
                file_path: Path = root / cell

                # Use the file name in the first cell as the subject name
                if subject is None:
                    subject_name = file_path.stem
                    if subject_name not in subjects:
                        subjects[subject_name] = self.add_subject(subject_name)
                    subject = subjects[subject_name]

                if file_type == 'shape':
                    shape_file = file_path
                    if not shape_file.exists():
                        raise Exception(f'Could not find shape file at "{shape_file}"')
                    data_type = shape_file_type(shape_file)
                    if data_type == Segmentation:
                        subject.add_segmentation(file=shape_file, anatomy_type=domain)
                    elif data_type == Mesh:
                        subject.add_mesh(file=shape_file, anatomy_type=domain)
                    found = True
                elif file_type == 'image':
                    image_file = file_path
                    if not image_file.exists():
                        raise Exception(f'Could not find image file at "{image_file}"')
                    subject.add_image(file=image_file, modality=domain)
                    found = True
        if not found:
            raise Exception('Did not find any shape or image files in data sheet')

    def download(self, path: Union[Path, str]):
        self.assert_remote()
        path = Path(path)
        dataset_file = None
        if self.file:
            dataset_file = self.file.download(path)
        for project in self.projects:
            project.download(path)
        if not dataset_file:
            return
        xls = load_workbook(str(dataset_file), read_only=True)
        for subject in self.subjects:
            row = next(entry for entry in xls['data'].values if subject.name == Path(entry[0]).stem)
            for segmentation in subject.segmentations:
                segmentation.file.download(path / Path(row[0]).parent)
            for mesh in subject.meshes:
                mesh.file.download(path / Path(row[0]).parent)
            for image in subject.images:
                image.file.download(path / Path(row[1]).parent)


class Subject(ApiModel):
    _endpoint = 'subjects'

    name: NonEmptyString
    dataset: Dataset

    @property
    def segmentations(self) -> Iterator[Segmentation]:
        self.assert_remote()
        return Segmentation.list(subject=self)

    @property
    def meshes(self) -> Iterator[Mesh]:
        self.assert_remote()
        return Mesh.list(subject=self)

    @property
    def images(self) -> Iterator[Image]:
        self.assert_remote()
        return Image.list(subject=self)

    def add_segmentation(self, file: Path, anatomy_type: str) -> Segmentation:
        return Segmentation(file=file, anatomy_type=anatomy_type, subject=self).create()

    def add_mesh(self, file: Path, anatomy_type: str) -> Mesh:
        return Mesh(file=file, anatomy_type=anatomy_type, subject=self).create()

    def add_image(self, file: Path, modality: str) -> Image:
        return Image(file=file, modality=modality, subject=self).create()


class Segmentation(ApiModel):
    _endpoint = 'segmentations'

    file: FileType[Literal['core.Segmentation.file']]
    anatomy_type: NonEmptyString
    subject: Subject


class Mesh(ApiModel):
    _endpoint = 'meshes'

    file: FileType[Literal['core.Mesh.file']]
    anatomy_type: NonEmptyString
    subject: Subject


class Image(ApiModel):
    _endpoint = 'images'

    file: FileType[Literal['core.Image.file']]
    modality: str
    subject: Subject


class Project(ApiModel):
    _endpoint = 'projects'

    file: FileType[Literal['core.Project.file']]
    keywords: str = ''
    description: str = ''
    dataset: Dataset

    @property
    def groomed_segmentations(self) -> Iterator[GroomedSegmentation]:
        self.assert_remote()
        return GroomedSegmentation.list(project=self)

    @property
    def groomed_meshes(self) -> Iterator[GroomedMesh]:
        self.assert_remote()
        return GroomedMesh.list(project=self)

    def add_groomed_segmentation(
        self,
        file: Path,
        segmentation: Segmentation,
        pre_cropping: Optional[Path] = None,
        pre_alignment: Optional[Path] = None,
    ) -> GroomedSegmentation:
        return GroomedSegmentation(
            file=file,
            segmentation=segmentation,
            project=self,
            pre_cropping=pre_cropping,
            pre_alignment=pre_alignment,
        ).create()

    def add_groomed_mesh(
        self,
        file: Path,
        mesh: Mesh,
        pre_cropping: Optional[Path] = None,
        pre_alignment: Optional[Path] = None,
    ) -> GroomedMesh:
        return GroomedMesh(
            file=file,
            mesh=mesh,
            project=self,
            pre_cropping=pre_cropping,
            pre_alignment=pre_alignment,
        ).create()

    @property
    def shape_models(self) -> Iterator[OptimizedShapeModel]:
        self.assert_remote()
        return OptimizedShapeModel.list(project=self)

    def add_shape_model(self, parameters: Dict[str, Any]) -> OptimizedShapeModel:
        return OptimizedShapeModel(
            project=self,
            parameters=parameters,
        ).create()

    def _load_project_spreadsheet(self) -> Project:
        file = self.file.path
        assert file  # should be guaranteed by assert_local

        root = file.parent
        xls = load_workbook(str(file), read_only=True)

        segmentations = {
            PurePath(segmentation.file.name).stem: segmentation
            for segmentation in self.dataset.segmentations
        }
        meshes = {PurePath(mesh.file.name).stem: mesh for mesh in self.dataset.meshes}

        # if 'optimize' not in xls or 'data' not in xls:
        #     raise Exception('`data` sheet not found')

        # shape_model = self._parse_optimize_sheet(xls['optimize'].values)
        # self._parse_data_sheet(segmentations, shape_model, xls['data'].values, root)

        for sheet_name in ['optimize', 'data']:
            if sheet_name not in xls:
                raise Exception(f'`{sheet_name}` sheet not found')

        shape_model = self._parse_optimize_sheet(xls['optimize'].values)
        self._parse_data_sheet(segmentations, meshes, shape_model, xls['data'], root)

        return self

    def _iter_data_sheet(
        self, sheet: Any, root: Path
    ) -> Iterator[Tuple[Path, Path, str, Path, Path, Path]]:
        # The old style ended in _file.  The newer style ends in an integer
        # (_1, _2, ...).  These can be in different orders
        # ##DWM::
        required = [
            'shape',
            'groomed',
            'alignment',
            'local_particles',
            'world_particles',
        ]
        optional = [
            'procrustes',
            'constraints',
        ]
        headers = next(sheet.values)
        if not any(all(k + suffix in headers for k in required) for suffix in ('_file', '_1')):
            raise Exception(
                'Unknown spreadsheet format - expected headers to include %r '
                'with a suffix of either _file or _1, found %r' % (required, headers)
            )
        for idx in itertools.count():
            suffix = '_file' if not idx else '_%d' % idx
            if not all(k + suffix in headers for k in required):
                if idx >= 1:
                    break
                continue
            col = {
                k: headers.index(k + suffix) for k in required + optional if k + suffix in headers
            }
            rows = sheet.values
            next(rows)  # skip header
            for rowrec in rows:
                row = {k: rowrec[col[k]] for k in col}
                if not row['shape']:
                    # It's possible to get XLSX files with empty rows where every column is None
                    continue

                shape_file = root / row['shape']
                groomed_file = root / row['groomed']
                local = root / row['local_particles']
                world = root / row['world_particles']
                alignment_file = row['alignment']
                constraints = (root / row['constraints']) if 'constraints' in row else None

                yield shape_file, groomed_file, alignment_file, local, world, constraints

    def _parse_optimize_sheet(self, sheet: Any) -> OptimizedShapeModel:
        expected = ('key', 'value')
        headers = next(sheet)
        if headers[: len(expected)] != expected:
            raise Exception(
                'Unknown spreadsheet format - expected headers to be %r, found %r'
                % (expected, headers[: len(expected)])
            )

        params: Dict[str, Union[str, float]] = {}
        for row in sheet:
            key, value = row
            try:
                value = float(value)
            except ValueError:
                pass
            params[key] = value

        return self.add_shape_model(params)

    def _parse_data_sheet(
        self,
        segmentations: Dict[str, Segmentation],
        meshes: Dict[str, Mesh],
        shape_model: OptimizedShapeModel,
        sheet: Any,
        root: Path,
    ):
        for (
            shape_file,
            groomed_file,
            alignment_file,
            local,
            world,
            constraints,
        ) in self._iter_data_sheet(sheet, root):
            groomed_segmentation = None
            groomed_mesh = None
            data_type = shape_file_type(shape_file)
            if data_type == Segmentation:
                segmentation = segmentations.get(shape_file.stem)
                if not segmentation:
                    raise Exception(f'Could not find segmentation for "{shape_file}"')
                groomed_segmentation = self.add_groomed_segmentation(
                    file=groomed_file,
                    segmentation=segmentation,
                )
            elif data_type == Mesh:
                mesh = meshes.get(shape_file.stem)
                if not mesh:
                    raise Exception(f'Could not find mesh for "{shape_file}"')
                groomed_mesh = self.add_groomed_mesh(
                    file=groomed_file,
                    mesh=mesh,
                )

            with TemporaryDirectory() as dir:
                transform = Path(dir) / f'{shape_file.stem}.transform'
                with transform.open('w') as f:
                    f.write(alignment_file)

                shape_model.add_particles(
                    world=world,
                    local=local,
                    transform=transform,
                    groomed_segmentation=groomed_segmentation,
                    groomed_mesh=groomed_mesh,
                    constraints=constraints,
                )

    def download(self, path: Union[Path, str]):
        path = Path(path)
        project_file = self.file.download(path)
        xls = load_workbook(str(project_file), read_only=True)
        sheet = xls['data']

        shape_model = next(self.shape_models)
        groomed_segmentations = {
            PurePath(gs.file.name).stem: gs for gs in self.groomed_segmentations
        }
        local_files = {PurePath(p.local.name).stem: p for p in shape_model.particles}

        # TODO: Do we detect if alignment_file (transform) is a path?
        for _, groomed_file, _, local, world, constraints in self._iter_data_sheet(sheet, path):
            gs = groomed_segmentations[groomed_file.stem]
            gs.file.download(groomed_file.parent)

            particles = local_files[local.stem]
            particles.local.download(local.parent)
            particles.world.download(world.parent)
            if particles.constraints and constraints:
                particles.constraints.download(constraints.parent)


class GroomedSegmentation(ApiModel):
    _endpoint = 'groomed-segmentations'

    file: FileType[Literal['core.GroomedSegmentation.file']]
    pre_cropping: Optional[FileType[Literal['core.GroomedSegmentation.pre_cropping']]] = None
    pre_alignment: Optional[FileType[Literal['core.GroomedSegmentation.pre_alignment']]] = None

    segmentation: Segmentation
    project: Project


class GroomedMesh(ApiModel):
    _endpoint = 'groomed-meshes'

    file: FileType[Literal['core.GroomedMesh.file']]
    pre_cropping: Optional[FileType[Literal['core.GroomedMesh.pre_cropping']]] = None
    pre_alignment: Optional[FileType[Literal['core.GroomedMesh.pre_alignment']]] = None

    mesh: Mesh
    project: Project


class OptimizedShapeModel(ApiModel):
    _endpoint = 'optimized-shape-models'

    project: Project
    parameters: Dict[str, Any]

    @property
    def particles(self) -> Iterator[OptimizedParticles]:
        return OptimizedParticles.list(shape_model=self)

    def add_particles(
        self,
        world: Path,
        local: Path,
        transform: Path,
        groomed_segmentation: Optional[GroomedSegmentation],
        groomed_mesh: Optional[GroomedMesh],
        constraints: Path,
    ) -> OptimizedParticles:
        return OptimizedParticles(
            world=world,
            local=local,
            transform=transform,
            shape_model=self,
            groomed_segmentation=groomed_segmentation,
            groomed_mesh=groomed_mesh,
            constraints=constraints,
        ).create()


class OptimizedParticles(ApiModel):
    _endpoint = 'optimized-particles'

    world: FileType[Literal['core.OptimizedParticles.world']]
    local: FileType[Literal['core.OptimizedParticles.local']]
    transform: FileType[Literal['core.OptimizedParticles.transform']]
    shape_model: OptimizedShapeModel
    groomed_segmentation: Optional[GroomedSegmentation]
    groomed_mesh: Optional[GroomedMesh]
    constraints: Optional[FileType[Literal['core.OptimizedParticles.constraints']]] = None


class OptimizedPCAModel(ApiModel):
    _endpoint = 'optimized-pca-model'

    mean_particles: FileType[Literal['core.OptimizedPCAModel.mean_particles']]
    pca_modes: FileType[Literal['core.OptimizedPCAModel.pca_modes']]
    eigen_spectrum: FileType[Literal['core.OptimizedPCAModel.eigen_spectrum']]
    shape_model: OptimizedShapeModel



import os

import pytest

from swcc import models
from swcc.api import swcc_session

from .factories import file_context as _file_context


@pytest.fixture(autouse=True)
def file_context():
    with _file_context() as d:
        yield d


@pytest.fixture
def session():
    base_url = os.getenv('DJANGO_BASE_URL', 'http://localhost:8000/api/v1')
    username = os.getenv('DJANGO_SUPERUSER_EMAIL', 'admin@noemail.com')
    password = os.getenv('DJANGO_SUPERUSER_PASSWORD', 'django-password')
    with swcc_session(base_url=base_url) as session:
        session.login(username, password)
        for dataset in models.Dataset.list():
            dataset.delete()

        for project in models.Project.list():
            project.delete()

        yield session



try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version  # type: ignore
import logging
from typing import Optional

from rich.logging import RichHandler

SWCC_CONFIG_PATH = __name__
SWCC_CONFIG_FILE = 'config'
FORMAT = '%(message)s'
logging.basicConfig(format=FORMAT, datefmt='[%X]', handlers=[RichHandler()])

__version__: Optional[str] = None
try:
    __version__ = version('swcc')
except PackageNotFoundError:
    # package is not installed
    pass



from __future__ import annotations

from collections import defaultdict
from contextlib import contextmanager
from typing import Any, Dict, List, Optional

import requests
from requests_toolbelt.sessions import BaseUrlSession
from s3_file_field_client import S3FileFieldClient
from urllib3.util.retry import Retry

from . import __version__

_session_stack: List[SwccSession] = []


def current_session():
    global _session_stack
    if not _session_stack:
        raise Exception('An active session context is required')

    return _session_stack[-1]


@contextmanager
def swcc_session(**kwargs):
    global _session_stack
    _session_stack.append(SwccSession(**kwargs))
    try:
        yield _session_stack[-1]
    finally:
        s = _session_stack.pop()
        s.close()


class SwccSession(BaseUrlSession):
    def __init__(
        self,
        base_url: str = 'https://app.shapeworks-cloud.org/api/v1',
        token: Optional[str] = None,
        **kwargs,
    ):
        base_url = f'{base_url.rstrip("/")}/'  # tolerate input with or without trailing slash
        super().__init__(base_url=base_url, **kwargs)

        self.cache: Dict[Any, Dict[int, Any]] = defaultdict(dict)
        retry = Retry()
        adapter = requests.adapters.HTTPAdapter(max_retries=retry)
        self.mount(base_url, adapter)

        self.headers.update(
            {
                'User-agent': f'swcc/{__version__}',
                'Accept': 'application/json',
            }
        )
        self.s3ff = S3FileFieldClient(f'{base_url}s3-upload/', self)
        if token:
            self.set_token(token)

    def set_token(self, token: str):
        self.headers['Authorization'] = f'Token {token}'

    def login(self, username: str, password: str) -> str:
        auth_url = f'{self.base_url.rstrip("/").replace("/api/v1", "")}/api-token-auth/'
        resp = requests.post(auth_url, data={'username': username, 'password': password})
        if resp.status_code != 200:
            raise Exception('Invalid username or password provided')
        token = resp.json()['token']
        self.set_token(token)
        return token



from __future__ import annotations

from datetime import datetime
import logging
import platform
import sys
import traceback

import click
from packaging.version import parse as parse_version
from pydantic import BaseModel
import requests
from requests.exceptions import RequestException

from . import SWCC_CONFIG_FILE, __version__
from .api import SwccSession
from .utils import get_config_value

logger = logging.getLogger(__name__)


class CliContext(BaseModel):
    session: SwccSession
    url: str
    json_output: bool

    class Config:
        arbitrary_types_allowed = True


def newer_version_available():
    if __version__ is None:
        return False

    this_version = parse_version(__version__)
    if this_version.is_devrelease:
        return False

    r = requests.get('https://pypi.org/pypi/swcc/json', timeout=(5, 5))
    r.raise_for_status()
    releases = [parse_version(v) for v in r.json()['releases'].keys()]
    for release in releases:
        if not (release.is_prerelease or release.is_devrelease) and release > this_version:
            return True
    return False


def session_response_handler(r: requests.Response, *args, **kwargs):
    if r.status_code in [401, 403]:
        click.echo(
            click.style(
                "You are attempting to perform an authorized operation but you aren't logged in.\n"  # noqa
                "Run 'swcc login' to continue.",
                fg='yellow',
            ),
            err=True,
        )
        sys.exit(1)


@click.group()
@click.option('--url', default='https://app.shapeworks-cloud.org/api/v1', envvar='SWCC_URL')
@click.option('--json', 'json_output', is_flag=True)
@click.option('-v', '--verbose', count=True)
@click.version_option()
@click.pass_context
def cli(ctx, url, json_output: bool, verbose: int):
    if verbose >= 2:
        logger.setLevel(logging.DEBUG)
    elif verbose >= 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARN)

    try:
        if newer_version_available():
            click.echo(
                click.style(
                    """There is a newer version of swcc available.
You must upgrade to the latest version before continuing.
""",
                    fg='yellow',
                ),
                err=True,
            )
            click.echo(click.style('pip install --upgrade swcc', fg='green'), err=True)
            sys.exit(1)
    except RequestException:
        click.echo(click.style('Failed to check for newer version of swcc:', fg='red'), err=True)
        raise

    token = get_config_value(SWCC_CONFIG_FILE, 'token')
    session = SwccSession(url)
    if token:
        session.set_token(token)
    session.hooks['response'] = session_response_handler
    ctx.obj = CliContext(
        session=session,
        url=url.rstrip('/'),
        json_output=json_output,
    )


@cli.group(name='groomed-dataset', short_help='get information about groomed datasets')
@click.pass_obj
def dataset(ctx):
    pass


def formatted_size(size, base=1024, unit='B'):
    if size < base:
        return f'{size} {unit}'
    units = ['', 'K', 'M', 'G', 'T']
    i = 0
    while i < 5 and size >= base:
        size /= base
        i += 1
    return f'{size:.2f} {units[i]}{unit}'


def main():
    try:
        cli()
    except Exception:
        click.echo(
            click.style(
                'The following unexpected error occurred while attempting your operation:\n',
                fg='red',
            ),
            err=True,
        )

        click.echo(traceback.format_exc(), err=True)

        click.echo(f'swcc:    v{__version__}', err=True)
        click.echo(f'python:  v{platform.python_version()}', err=True)
        click.echo(f'time:    {datetime.utcnow().isoformat()}', err=True)
        click.echo(f'os:      {platform.platform()}', err=True)
        click.echo(f'command: swcc {" ".join(sys.argv[1:])}\n', err=True)

        click.echo(
            click.style(
                'This is a bug in swcc and should be reported. You can open an issue below: ',
                fg='yellow',
            ),
            err=True,
        )
        click.echo(
            'https://github.com/girder/shapeworks-cloud/issues/new',
            err=True,
        )


if __name__ == '__main__':
    main()



from pathlib import Path

from setuptools import find_packages, setup

readme_file = Path(__file__).parent / 'README.md'
if readme_file.exists():
    with readme_file.open() as f:
        long_description = f.read()
else:
    # When this is first installed in development Docker, README.md is not available
    long_description = ''

setup(
    name='swcc',
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    keywords='',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python',
    ],
    python_requires='>=3.8',
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'click<8',  # celery depends on click<8, pin for now
        'click-pathlib',
        'django-s3-file-field-client',
        # if you override and install in python 3.7
        'importlib_metadata ; python_version < "3.8"',
        'openpyxl',
        'packaging',
        'pydantic',
        'pyxdg',
        'requests',
        'requests-toolbelt',
        'rich',
        'toml',
        'tqdm',
        'typing-extensions ; python_version < "3.8"',
    ],
    entry_points={'console_scripts': ['swcc=swcc.cli:main']},
    extras_require={'dev': ['factory_boy', 'ipython', 'tox']},
)



from pathlib import Path
from tempfile import TemporaryDirectory

from factory import Factory
import pytest
import requests

from swcc import models

from . import factories

_factories = [
    f
    for f in factories.__dict__.values()
    if isinstance(f, type) and issubclass(f, Factory) and f is not Factory
]


def test_api_common(session):
    dataset1 = factories.DatasetFactory().create()
    dataset2 = factories.DatasetFactory().create()

    other = models.Dataset.from_id(dataset1.id)
    assert other.name == dataset1.name

    assert {d.id for d in models.Dataset.list()} == {dataset1.id, dataset2.id}

    dataset1.delete()
    assert {d.id for d in models.Dataset.list()} == {dataset2.id}


def test_file_download_api(session):
    with TemporaryDirectory() as d:
        directory = Path(d)
        file = directory / 'test_file.nrrd'
        with file.open('w') as f:
            f.write('file content')
        segmentation = factories.SegmentationFactory(file=file).create()
        segmentation = models.Segmentation.from_id(segmentation.id)

        assert file.name in segmentation.file.url

        d1 = directory / '1'
        d1.mkdir()
        segmentation.file.download(d1)

        download_file = d1 / 'test_file.nrrd'
        assert download_file.exists()
        with download_file.open('r') as f:
            assert f.read() == 'file content'

        d2 = directory / '2'
        d2.mkdir()
        files = list(segmentation.download_files(d2))
        assert files == [d2 / 'test_file.nrrd']


@pytest.mark.parametrize('factory', _factories)
def test_model_crud(session, factory):
    obj = factory().create()
    model = obj.__class__

    assert model.from_id(obj.id).id == obj.id
    assert obj.id in {o.id for o in model.list()}

    obj.delete()
    assert obj.id not in {o.id for o in model.list()}


def test_get_dataset_by_name(session):
    factories.DatasetFactory(name='dataset1').create()
    factories.DatasetFactory(name='dataset2').create()

    dataset = models.Dataset.from_name('dataset1')
    assert dataset and dataset.name == 'dataset1'

    dataset = models.Dataset.from_name('dataset2')
    assert dataset and dataset.name == 'dataset2'

    dataset = models.Dataset.from_name('dataset3')
    assert dataset is None


def test_dataset_force_create_overwrite(session):
    old_dataset = factories.DatasetFactory(name='dataset1').create()
    assert models.Dataset.from_id(old_dataset.id)

    new_dataset = models.Dataset(
        name='dataset1',
        license=old_dataset.license,
        description=old_dataset.description,
        acknowledgement=old_dataset.acknowledgement,
    ).force_create()
    new_dataset.assert_remote()

    # The new dataset should now exist on the server
    assert models.Dataset.from_name('dataset1') == new_dataset
    # The old dataset should no longer exist on the server
    with pytest.raises(requests.exceptions.HTTPError):
        # Gotta clear the cache first
        session.cache[models.Dataset] = {}
        old = models.Dataset.from_id(old_dataset.id)
        assert old


def test_dataset_force_create_no_existing(session):
    dataset = models.Dataset(
        name='dataset1',
        license='license',
        description='description',
        acknowledgement='acknowledgement',
    ).force_create()
    dataset.assert_remote()

    assert models.Dataset.from_name('dataset1') == dataset


@pytest.mark.parametrize(
    'version',
    [0, 1, 9, 10, 99],
)
def test_dataset_force_create_backup(session, version):
    old_name = f'dataset-v{version}'
    new_name = f'dataset-v{version+1}'
    old_dataset = factories.DatasetFactory(name=old_name).create()
    assert models.Dataset.from_id(old_dataset.id)

    new_dataset = models.Dataset(
        name=old_name,
        license=old_dataset.license,
        description=old_dataset.description,
        acknowledgement=old_dataset.acknowledgement,
    ).force_create(backup=True)
    new_dataset.assert_remote()

    # The old dataset should still exist on the server
    assert models.Dataset.from_name(old_name) == old_dataset
    # The new dataset should also exist on the server
    assert models.Dataset.from_name(new_name) == new_dataset
    # The new dataset should have its name changed in place
    assert new_dataset.name == new_name


def test_dataset_force_create_backup_no_version_suffix(session):
    old_dataset = factories.DatasetFactory(name='dataset1').create()
    assert models.Dataset.from_id(old_dataset.id)

    new_dataset = models.Dataset(
        name='dataset1',
        license=old_dataset.license,
        description=old_dataset.description,
        acknowledgement=old_dataset.acknowledgement,
    ).force_create(backup=True)
    new_dataset.assert_remote()

    # The old dataset should still exist on the server
    assert models.Dataset.from_name('dataset1') == old_dataset
    # The new dataset should also exist on the server
    assert models.Dataset.from_name('dataset1-v1') == new_dataset


def test_dataset_force_create_backup_multiple_conflicts(session):
    old_dataset = factories.DatasetFactory(name='dataset').create()
    factories.DatasetFactory(name='dataset-v1').create()
    factories.DatasetFactory(name='dataset-v2').create()
    factories.DatasetFactory(name='dataset-v3').create()
    assert models.Dataset.from_id(old_dataset.id)

    new_dataset = models.Dataset(
        name='dataset',
        license=old_dataset.license,
        description=old_dataset.description,
        acknowledgement=old_dataset.acknowledgement,
    ).force_create(backup=True)
    new_dataset.assert_remote()

    # The new dataset should cascade up until it reaches dataset-v4
    assert models.Dataset.from_name('dataset-v4') == new_dataset
